<?php

namespace KnitPay\Extensions\KnitPayPaymentButton;

use Pronamic\WordPress\Html\Element;
use Pronamic\WordPress\Money\Currency;
use Pronamic\WordPress\Money\Money;
use Pronamic\WordPress\Pay\Plugin;
use Pronamic\WordPress\Pay\Util;
use Pronamic\WordPress\Pay\Payments\Payment;

/**
 * Title: Knit Pay - Payment Button Gateway
 * Description:
 * Copyright: 2020-2023 Knit Pay
 * Company: Knit Pay
 *
 * @author  knitpay
 * @since   8.75.0.0
 */

/**
 * Prevent loading this file directly
 */
defined( 'ABSPATH' ) || exit();

class Gateway {

	/**
	 * Bootstrap
	 */
	public function __construct() {
		// Initiate Payment Button Block.
		if ( function_exists( 'register_block_type' ) ) {
			$this->blocks_module = new PaymentButtonBlock();
			$this->blocks_module->setup();
		}

		add_action( 'wp_ajax_nopriv_knit_pay_payment_button_submit', [ $this, 'ajax_payment_button_submit' ] );
		add_action( 'wp_ajax_knit_pay_payment_button_submit', [ $this, 'ajax_payment_button_submit' ] );
	}

	/**
	 * Input Field.
	 *
	 * @param array $args Arguments.
	 * @return void
	 */
	public function input_field( $args ) {
		$args['id']   = $args['label_for'];
		$args['name'] = $args['label_for'];

		$element = new Element( 'input', $args );
		$element->output();

		if ( isset( $args['description'] ) ) {
			printf(
				'<p class="pronamic-pay-description description">%s</p>',
	            // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
				$args['description']
			);
		}
	}

	/**
	 * Input page.
	 *
	 * @param array $args Arguments.
	 * @return void
	 */
	public function select_configuration( $args ) {
		$args['id']   = $args['label_for'];
		$args['name'] = $args['label_for'];

		$configurations    = Plugin::get_config_select_options();
		$configurations[0] = __( '— Default Gateway —', 'knit-pay-lang' );

		$configuration_options              = [];
		$configuration_options[]['options'] = $configurations;

		printf(
			'<select %s>%s</select>',
	        // @codingStandardsIgnoreStart
	        Util::array_to_html_attributes( $args ),
	        Util::select_options_grouped( $configuration_options, get_option( 'pronamic_pay_config_id' ) )
	        // @codingStandardsIgnoreEnd
		);

		if ( isset( $args['description'] ) ) {
			printf(
				'<p class="pronamic-pay-description description">%s</p>',
	            // phpcs:ignore WordPress.Security.EscapeOutput.OutputNotEscaped
				$args['description']
			);
		}
	}

	public function ajax_payment_button_submit() {
		$amount              = filter_input( INPUT_POST, 'amount', FILTER_SANITIZE_STRING );
		$currency            = filter_input( INPUT_POST, 'currency', FILTER_SANITIZE_STRING );
		$payment_description = filter_input( INPUT_POST, 'payment_description', FILTER_SANITIZE_STRING );
		$config_id           = filter_input( INPUT_POST, 'config_id', FILTER_SANITIZE_STRING );
		$nonce_action        = "knit_pay_payment_button|{$amount}|{$currency}|{$payment_description}|{$config_id}";

		if ( ! wp_verify_nonce( filter_input( INPUT_POST, 'knit_pay_nonce', FILTER_SANITIZE_STRING ), $nonce_action ) ) {
			echo wp_json_encode(
				[
					'status'    => 'error',
					'error_msg' => __( 'Nonce Missmatch!', 'knit-pay' ),
				]
			);
			exit;
		}

		if ( empty( $amount ) ) {
			$amount = 0;
		}

		$payment_method = 'knit_pay';
		$current_user   = wp_get_current_user();

		// Use default gateway if no configuration has been set.
		if ( empty( $config_id ) ) {
			$config_id = get_option( 'pronamic_pay_config_id' );
		}

		$gateway = Plugin::get_gateway( $config_id );

		if ( ! $gateway ) {
			return false;
		}

		/**
		 * Build payment.
		 */
		$payment = new Payment();

		$payment->source   = 'knit-pay-payment-button';
		$payment->order_id = uniqid();
		$payment->source   = uniqid();

		$payment->set_description( Helper::get_description( $payment, $payment_description ) );

		$payment->title = Helper::get_title( $payment );

		// Customer.
		$payment->set_customer( Helper::get_customer( $current_user ) );

		// Address.
		$payment->set_billing_address( Helper::get_address( $current_user ) );

		// Currency.
		$currency = Currency::get_instance( $currency );

		// Amount.
		$payment->set_total_amount( new Money( $amount, $currency ) );

		// Method.
		$payment->set_payment_method( $payment_method );

		// Configuration.
		$payment->config_id = $config_id;

		try {
			$payment = Plugin::start_payment( $payment );

			// Execute a redirect.
			echo wp_json_encode(
				[
					'status'       => 'success',
					'redirect_url' => $payment->get_pay_redirect_url(),
				]
			);
			exit;
		} catch ( \Exception $e ) {
			echo wp_json_encode(
				[
					'status'    => 'error',
					'error_msg' => $e->getMessage(),
				]
			);
			exit;
		}
	}

	public static function instance() {
		return new self();
	}
}
